#include <iostream> 
#include <fstream>
#include <utility>
#include <string>
#include <boost/graph/graph_traits.hpp>
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/random.hpp>
#include <boost/random.hpp>
#include <boost/graph/astar_search.hpp>
#include <vector>
#include <ctime>
#include <thread>
#include <map>
#include <mutex>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <sstream>
#include <libpq-fe.h>
#include "RoadNetwork.hpp"
#include "OD.hpp"
#include "Raster.hpp"
#ifndef ROUTING_H
#define ROUTING_H

using namespace std;

extern int NUM_THREADS;
extern int NUM_INCREMENTS;
extern bool DEBUG;
extern mutex vol_mutex;

int singleRoute(RoadNetwork g, 
    NodeMap* locations, 
    int start, 
    int goal 
    )
{

    distance_heuristic<RoadNetwork, float, NodeMap> heuristic(*locations, goal);
    vector<int> p(boost::num_vertices(g));
    vector<cost> d(boost::num_vertices(g));

    try 
    {
        // call astar named parameter interface
        boost::astar_search_tree(g, 
            start,
            heuristic,
            boost::weight_map(boost::get(&Edge::travel_time, g)).predecessor_map(boost::make_iterator_property_map(p.begin(), boost::get(boost::vertex_index, g))).
            distance_map(boost::make_iterator_property_map(d.begin(), boost::get(boost::vertex_index, g))).
            visitor(astar_goal_visitor<int>(goal))
            );
    } catch(found_goal fg) 
    { // found a path to the goal
        list<int> shortest_path;
        for(int v = goal;; v = p[v]) 
        {
            shortest_path.push_front(v);
            if(p[v] == v)
                break;
        }

    cout << "Shortest path from " << start << " to " << goal << ": LINESTRING (";
    list<int>::iterator spi = shortest_path.begin();
    cout << g[start].lon << ' ' << g[start].lat ;
    for(++spi; spi != shortest_path.end(); ++spi)
        cout << "," << g[*spi].lon << " " << g[*spi].lat;
    cout << ")" << endl << "Total travel time: " << d[goal] << endl;
    
    }
    return 0;
}

int batchRoute( RoadNetwork* g, 
    EdgeMap* volume, 
    EdgeTractMap* tractMap,
    OD* od, 
    NodeMap* locations, 
    int b,
    Raster<int>* raster_map
    )
{
    // get the batch of ODs from the large vector
    OD batch = (*od).getBatch(b);
    if(DEBUG)
    {
        cout << "Starting batch: " << b << endl;
        cout << "size of batch: " << batch.pairs.size() << endl;
        cout << "first tuple: " << batch.pairs[0].s << "," << batch.pairs[0].t << "," << batch.pairs[0].flow << endl;
    }

    clock_t startTime = clock();
    int found = 0;
    // route all ods
    for( int n=0; n < batch.pairs.size(); n++)
    {
        if (DEBUG && (n%10000==0))
        { 
            cout << "Batch " << b << " : " << double( clock() - startTime ) / (double)CLOCKS_PER_SEC / (double)NUM_THREADS << " seconds | ";
            cout << n << " of " << batch.pairs.size() << " | ";
            cout << "found: " << found << endl;
            found = 0;
            startTime = clock();
        }
        int start = batch.pairs[n].s;
        int goal = batch.pairs[n].t;
        int startTract = (*raster_map).inPoly((*g)[start].lon, (*g)[start].lat);
        int goalTract = (*raster_map).inPoly((*g)[goal].lon, (*g)[goal].lat);

        distance_heuristic<RoadNetwork, float, NodeMap> heuristic(*locations, goal);
    	vector<int> p(boost::num_vertices(*g));
    	vector<cost> d(boost::num_vertices(*g));

        try 
        {
    	    // call astar named parameter interface
    	    boost::astar_search_tree(*g, 
    	    	start,
    	    	heuristic,
    	    	boost::weight_map(boost::get(&Edge::travel_time, *g)).predecessor_map(boost::make_iterator_property_map(p.begin(), boost::get(boost::vertex_index, *g))).
           		distance_map(boost::make_iterator_property_map(d.begin(), boost::get(boost::vertex_index, *g))).
           		visitor(astar_goal_visitor<int>(goal))
    	    	);
        } catch(found_goal fg) 
        { // found a path to the goal
            found++;
    	    for(int v = goal;; v = p[v]) 
            {
    	    	if(p[v] == v)
    	    		break;
                // update volume on edges traveled.
                edge_descriptor e = boost::edge(p[v],v,*g).first;
                vol_mutex.lock();
                // update volumes and tract origins
                (*volume)[e] += batch.pairs[n].flow;
                (*tractMap)[e][startTract].first += batch.pairs[n].flow;
                (*tractMap)[e][goalTract].second += batch.pairs[n].flow;
                vol_mutex.unlock();
            }
        }
	}
    if (DEBUG)
        cout << "Batch " << b << " is finished." << endl;
    return 0;
}
#endif
